@@ -636,6 +636,8 @@ DhtProxyClient::queryProxyInfo(const Sp<InfoState>& infoState, const Sp<http::Re
         logger_->d("[proxy:client] [status] query ipv%i info", family == AF_INET ? 4 : 6);
     try {
         auto request = std::make_shared<http::Request>(httpContext_, resolver, family);
+        if (serverCertificate_)
+            request->set_certificate_authority(serverCertificate_);
         auto reqid = request->id();
         request->set_method(restinio::http_method_get());
         setHeaderFields(*request);
@@ -1162,6 +1164,7 @@ DhtProxyClient::pushNotificationReceived(const std::map<std::string, std::string
         statusIpv4_ = NodeStatus::Connected;
         statusIpv6_ = NodeStatus::Connected;
     }
+    auto launchLoop = false;
     try {
         auto sessionId = notification.find("s");
         if (sessionId != notification.end() and sessionId->second != pushSessionId_) {
@@ -1238,7 +1241,7 @@ DhtProxyClient::pushNotificationReceived(const std::map<std::string, std::string
                                 opstate->stop = true;
                         });
                     }
-                    loopSignal_();
+                    launchLoop = true;
                 }
             }
         }
@@ -1246,6 +1249,8 @@ DhtProxyClient::pushNotificationReceived(const std::map<std::string, std::string
         if (logger_)
             logger_->e("[proxy:client] [push] receive error: %s", e.what());
     }
+    if (launchLoop)
+        loopSignal_();
 #else
     (void) notification;
 #endif
