@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2014-2018 Savoir-faire Linux Inc.
+ *  Copyright (C) 2014-2020 Savoir-faire Linux Inc.
  *  Author(s) : Adrien Béraud <adrien.beraud@savoirfairelinux.com>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -65,10 +65,16 @@ struct Dht::SearchNode {
         ValueCache cache;
         Sp<Scheduler::Job> cacheExpirationJob {};
         Sp<net::Request> req {};
-        CachedListenStatus(ValueStateCallback&& cb) : cache(std::forward<ValueStateCallback>(cb)) {}
-        CachedListenStatus(CachedListenStatus&&) = default;
+        Tid socketId {0};
+        CachedListenStatus(ValueStateCallback&& cb, SyncCallback scb, Tid sid)
+         : cache(std::forward<ValueStateCallback>(cb), std::forward<SyncCallback>(scb)), socketId(sid) {}
+        CachedListenStatus(CachedListenStatus&&) = delete;
         CachedListenStatus(const CachedListenStatus&) = delete;
-        CachedListenStatus& operator=(const CachedListenStatus&) = delete;
+        ~CachedListenStatus() {
+            if (socketId and req and req->node) {
+                req->node->closeSocket(socketId);
+            }
+        }
     };
     using NodeListenerStatus = std::map<Sp<Query>, CachedListenStatus>;
 
@@ -85,15 +91,15 @@ struct Dht::SearchNode {
 
     Blob token {};                                 /* last token the node sent to us after a get request */
     time_point last_get_reply {time_point::min()}; /* last time received valid token */
+    Sp<Scheduler::Job> syncJob {};
     bool candidate {false};                        /* A search node is candidate if the search is/was synced and this
                                                       node is a new candidate for inclusion. */
-    Sp<Scheduler::Job> syncJob {};
 
     SearchNode() : node() {}
     SearchNode(const SearchNode&) = delete;
-    SearchNode(SearchNode&&) = default;
+    SearchNode(SearchNode&&) = delete;
     SearchNode& operator=(const SearchNode&) = delete;
-    SearchNode& operator=(SearchNode&&) = default;
+    SearchNode& operator=(SearchNode&&) = delete;
 
     SearchNode(const Sp<Node>& node) : node(node) {}
     ~SearchNode() {
@@ -225,6 +231,13 @@ struct Dht::SearchNode {
         }
     }
 
+    void onListenSynced(const Sp<Query>& q, bool synced = true) {
+        auto l = listenStatus.find(q);
+        if (l != listenStatus.end()) {
+            l->second.cache.onSynced(synced);
+        }
+    }
+
     void expireValues(const Sp<Query>& q, Scheduler& scheduler) {
         auto l = listenStatus.find(q);
         if (l != listenStatus.end()) {
@@ -255,7 +268,7 @@ struct Dht::SearchNode {
      * @return true if there exists an expired request, else false.
      */
     static bool pending(const SyncStatus& status) {
-        return std::find_if(status.begin(), status.end(),
+        return std::find_if(status.cbegin(), status.cend(),
             [](const SyncStatus::value_type& r){
                 return r.second and r.second->pending();
             }) != status.end();
@@ -285,26 +298,26 @@ struct Dht::SearchNode {
         acked.clear();
     }
 
-    bool isListening(time_point now) const {
+    inline bool isListening(time_point now, duration listen_expire) const {
         auto ls = listenStatus.begin();
         for ( ; ls != listenStatus.end() ; ++ls) {
-            if (isListening(now, ls)) {
+            if (isListening(now, ls, listen_expire)) {
                 break;
             }
         }
         return ls != listenStatus.end();
     }
-    bool isListening(time_point now, const Sp<Query>& q) const {
+    inline bool isListening(time_point now, const Sp<Query>& q, duration listen_expire) const {
         const auto& ls = listenStatus.find(q);
         if (ls == listenStatus.end())
             return false;
         else
-            return isListening(now, ls);
+            return isListening(now, ls, listen_expire);
     }
-    bool isListening(time_point now, NodeListenerStatus::const_iterator listen_status) const {
+    inline bool isListening(time_point now, NodeListenerStatus::const_iterator listen_status, duration listen_expire) const {
         if (listen_status == listenStatus.end() or not listen_status->second.req)
             return false;
-        return listen_status->second.req->reply_time + LISTEN_EXPIRE_TIME > now;
+        return listen_status->second.req->reply_time + listen_expire > now;
     }
     void cancelListen() {
         for (const auto& status : listenStatus)
@@ -337,12 +350,14 @@ struct Dht::SearchNode {
      * Assuming the node is synced, should the "listen" request with Query q be
      * sent to this node now ?
      */
-    time_point getListenTime(const Sp<Query>& q) const {
-        auto listen_status = listenStatus.find(q);
-        if (listen_status == listenStatus.end() or not listen_status->second.req)
+    time_point getListenTime(const decltype(listenStatus)::const_iterator listen_status, duration listen_expire) const {
+        if (listen_status == listenStatus.cend() or not listen_status->second.req)
             return time_point::min();
         return listen_status->second.req->pending() ? time_point::max() :
-            listen_status->second.req->reply_time + LISTEN_EXPIRE_TIME - REANNOUNCE_MARGIN;
+            listen_status->second.req->reply_time + listen_expire - REANNOUNCE_MARGIN;
+    }
+    time_point getListenTime(const Sp<Query>& q, duration listen_expire) const {
+        return getListenTime(listenStatus.find(q), listen_expire);
     }
 
     /**
@@ -368,7 +383,7 @@ struct Dht::Search {
 
     bool expired {false};              /* no node, or all nodes expired */
     bool done {false};                 /* search is over, cached for later */
-    std::vector<SearchNode> nodes {};
+    std::vector<std::unique_ptr<SearchNode>> nodes {};
 
     /* pending puts */
     std::vector<Announce> announce {};
@@ -377,6 +392,11 @@ struct Dht::Search {
     std::multimap<time_point, Get> callbacks {};
 
     /* listeners */
+    struct SearchListener {
+        Sp<Query> query;
+        ValueCallback get_cb;
+        SyncCallback sync_cb;
+    };
     std::map<size_t, SearchListener> listeners {};
     size_t listener_token = 1;
 
@@ -403,17 +423,17 @@ struct Dht::Search {
     bool insertNode(const Sp<Node>& n, time_point now, const Blob& token={});
 
     SearchNode* getNode(const Sp<Node>& n) {
-        auto srn = std::find_if(nodes.begin(), nodes.end(), [&](SearchNode& sn) {
-            return n == sn.node;
+        auto srn = std::find_if(nodes.begin(), nodes.end(), [&](std::unique_ptr<SearchNode>& sn) {
+            return n == sn->node;
         });
-        return (srn == nodes.end()) ? nullptr : &(*srn);
+        return (srn == nodes.end()) ? nullptr : (*srn).get();
     }
 
     /* number of concurrent sync requests */
     unsigned currentlySolicitedNodeCount() const {
         unsigned count = 0;
         for (const auto& n : nodes)
-            if (not n.isBad() and n.pendingGet())
+            if (not n->isBad() and n->pendingGet())
                 count++;
         return count;
     }
@@ -447,12 +467,12 @@ struct Dht::Search {
      */
     void setDone(const Get& get) {
         for (auto& n : nodes) {
-            auto pqs = n.pagination_queries.find(get.query);
-            if (pqs != n.pagination_queries.cend()) {
+            auto pqs = n->pagination_queries.find(get.query);
+            if (pqs != n->pagination_queries.cend()) {
                 for (auto& pq : pqs->second)
-                    n.getStatus.erase(pq);
+                    n->getStatus.erase(pq);
             }
-            n.getStatus.erase(get.query);
+            n->getStatus.erase(get.query);
         }
         if (get.done_cb)
             get.done_cb(true, getNodes());
@@ -465,33 +485,32 @@ struct Dht::Search {
      */
     void setDone() {
         for (auto& n : nodes) {
-            n.getStatus.clear();
-            n.listenStatus.clear();
-            n.acked.clear();
+            n->getStatus.clear();
+            n->listenStatus.clear();
+            n->acked.clear();
         }
         done = true;
     }
 
     bool isAnnounced(Value::Id id) const;
-    bool isListening(time_point now) const;
+    bool isListening(time_point now, duration exp) const;
 
-    void get(Value::Filter f, const Sp<Query>& q, const QueryCallback& qcb, const GetCallback& gcb, const DoneCallback& dcb, Scheduler& scheduler) {
+    void get(const Value::Filter& f, const Sp<Query>& q, const QueryCallback& qcb, const GetCallback& gcb, const DoneCallback& dcb, Scheduler& scheduler) {
         if (gcb or qcb) {
-            const auto& now = scheduler.time();
-            callbacks.emplace(now, Get { now, f, q, qcb, gcb, dcb });
-            auto values = cache.get(f);
-            if (not values.empty())
-                gcb(values);
-            scheduler.edit(nextSearchStep, now);
+            if (not cache.get(f, q, gcb, dcb)) {
+                const auto& now = scheduler.time();
+                callbacks.emplace(now, Get { now, f, q, qcb, gcb, dcb });
+                scheduler.edit(nextSearchStep, now);
+            }
         }
     }
 
-    size_t listen(ValueCallback cb, Value::Filter f, const Sp<Query>& q, Scheduler& scheduler) {
+    size_t listen(const ValueCallback& cb, const Value::Filter& f, const Sp<Query>& q, Scheduler& scheduler) {
         //DHT_LOG.e(id, "[search %s IPv%c] listen", id.toString().c_str(), (af == AF_INET) ? '4' : '6');
-        return cache.listen(cb, q, f, [&](const Sp<Query>& q, ValueCallback vcb){
+        return cache.listen(cb, q, f, [&](const Sp<Query>& q, ValueCallback vcb, SyncCallback scb){
             done = false;
             auto token = ++listener_token;
-            listeners.emplace(token, SearchListener{q, f, vcb});
+            listeners.emplace(token, SearchListener{q, vcb, scb});
             scheduler.edit(nextSearchStep, scheduler.time());
             return token;
         });
@@ -510,9 +529,9 @@ struct Dht::Search {
                     }
                     for (auto& sn : nodes) {
                         if (listeners.empty())
-                            sn.cancelListen();
+                            sn->cancelListen();
                         else if (query)
-                            sn.cancelListen(query);
+                            sn->cancelListen(query);
                     }
                 });
                 scheduler.edit(opExpirationJob, nextExpire);
@@ -520,18 +539,92 @@ struct Dht::Search {
         scheduler.edit(opExpirationJob, cache.getExpiration());
     }
 
+    std::vector<Sp<Value>> getPut() const {
+        std::vector<Sp<Value>> ret;
+        ret.reserve(announce.size());
+        for (const auto& a : announce)
+            ret.push_back(a.value);
+        return ret;
+    }
+
+    Sp<Value> getPut(Value::Id vid) const {
+        for (auto& a : announce) {
+            if (a.value->id == vid)
+                return a.value;
+        }
+        return {};
+    }
+
+    bool cancelPut(Value::Id vid) {
+        bool canceled {false};
+        for (auto it = announce.begin(); it != announce.end();) {
+            if (it->value->id == vid) {
+                canceled = true;
+                it = announce.erase(it);
+            }
+            else
+                ++it;
+        }
+        for (auto& n : nodes) {
+            auto ackIt = n->acked.find(vid);
+            if (ackIt != n->acked.end()) {
+                if (ackIt->second.first)
+                    ackIt->second.first->cancel();
+                n->acked.erase(ackIt);
+            }
+        }
+        return canceled;
+    }
+
+    void put(const Sp<Value>& value, DoneCallback callback, time_point created, bool permanent) {
+        done = false;
+        expired = false;
+        auto a_sr = std::find_if(announce.begin(), announce.end(), [&](const Announce& a){
+            return a.value->id == value->id;
+        });
+        if (a_sr == announce.end()) {
+            announce.emplace_back(Announce {permanent, value, created, callback});
+            for (auto& n : nodes) {
+                n->probe_query.reset();
+                n->acked[value->id].first.reset();
+            }
+        } else {
+            a_sr->permanent = permanent;
+            a_sr->created = created;
+            if (a_sr->value != value) {
+                a_sr->value = value;
+                for (auto& n : nodes) {
+                    n->acked[value->id].first.reset();
+                    n->probe_query.reset();
+                }
+            }
+            if (isAnnounced(value->id)) {
+                if (a_sr->callback)
+                    a_sr->callback(true, {});
+                a_sr->callback = {};
+                if (callback)
+                    callback(true, {});
+                return;
+            } else {
+                if (a_sr->callback)
+                    a_sr->callback(false, {});
+                a_sr->callback = callback;
+            }
+        }
+    }
+
     /**
      * @return The number of non-good search nodes.
      */
     unsigned getNumberOfBadNodes() const {
-        return std::count_if(nodes.begin(), nodes.end(), [](const SearchNode& sn) {
-            return sn.isBad();
+        return std::count_if(nodes.begin(), nodes.end(), [](const std::unique_ptr<SearchNode>& sn) {
+            return sn->isBad();
         });
     }
     unsigned getNumberOfConsecutiveBadNodes() const {
         unsigned count = 0;
-        std::find_if(nodes.begin(), nodes.end(), [&count](const SearchNode& sn) {
-            if (not sn.isBad())
+        std::find_if(nodes.begin(), nodes.end(), [&count](const std::unique_ptr<SearchNode>& sn) {
+            if (not sn->node->isExpired())
                 return true;
             ++count;
             return false;
@@ -550,7 +643,7 @@ struct Dht::Search {
      */
     bool removeExpiredNode(const time_point& now) {
         for (auto e = nodes.cend(); e != nodes.cbegin();) {
-            const Node& n = *(--e)->node;
+            const Node& n = *(*(--e))->node;
             if (n.isRemovable(now)) {
                 //std::cout << "Removing expired node " << n.id << " from IPv" << (af==AF_INET?'4':'6') << " search " << id << std::endl;
                 nodes.erase(e);
@@ -622,8 +715,14 @@ struct Dht::Search {
         });
         // remove acked for cleared annouces
         for (auto it = announced; it != announce.end(); ++it) {
-            for (auto& n : nodes)
-                n.acked.erase(it->value->id);
+            for (auto& n : nodes) {
+                auto ackIt = n->acked.find(it->value->id);
+                if (ackIt != n->acked.end()) {
+                    if (ackIt->second.first)
+                        ackIt->second.first->cancel();
+                    n->acked.erase(ackIt);
+                }
+            }
         }
         announce.erase(announced, announce.end());
     }
@@ -656,19 +755,18 @@ Dht::Search::insertNode(const Sp<Node>& snode, time_point now, const Blob& token
     auto n = nodes.end();
     while (n != nodes.begin()) {
         --n;
-        if (n->node == snode) {
+        if ((*n)->node == snode) {
             found = true;
             break;
         }
 
         /* Node not found. We could insert it after this one. */
-        if (id.xorCmp(nid, n->node->id) > 0) {
+        if (id.xorCmp(nid, (*n)->node->id) > 0) {
             ++n;
             break;
         }
     }
 
-    bool new_search_node = false;
     if (not found) {
         // find if and where to trim excessive nodes
         auto t = nodes.cend();
@@ -686,15 +784,16 @@ Dht::Search::insertNode(const Sp<Node>& snode, time_point now, const Blob& token
             full = nodes.size() - bad >=  SEARCH_NODES;
             while (std::distance(nodes.cbegin(), t) - bad >  SEARCH_NODES) {
                 --t;
-                if (t->isBad())
+                if ((*t)->isBad())
                     bad--;
             }
         }
 
         if (full) {
+            bool addNode = n < t;
             if (t != nodes.cend())
                 nodes.resize(std::distance(nodes.cbegin(), t));
-            if (n >= t)
+            if (not addNode)
                 return false;
         }
 
@@ -702,9 +801,8 @@ Dht::Search::insertNode(const Sp<Node>& snode, time_point now, const Blob& token
         if (nodes.empty()) {
             step_time = time_point::min();
         }
-        n = nodes.insert(n, SearchNode(snode));
+        n = nodes.emplace(n, std::make_unique<SearchNode>(snode));
         node.setTime(now);
-        new_search_node = true;
         if (node.isExpired()) {
             if (not expired)
                 bad++;
@@ -714,21 +812,21 @@ Dht::Search::insertNode(const Sp<Node>& snode, time_point now, const Blob& token
         }
 
         while (nodes.size() - bad >  SEARCH_NODES) {
-            if (not expired and nodes.back().isBad())
+            if (not expired and nodes.back()->isBad())
                 bad--;
             nodes.pop_back();
         }
     }
     if (not token.empty()) {
-        n->candidate = false;
-        n->last_get_reply = now;
+        (*n)->candidate = false;
+        (*n)->last_get_reply = now;
         if (token.size() <= 64)
-            n->token = token;
+            (*n)->token = token;
         expired = false;
     }
-    if (new_search_node)
+    if (not found)
         removeExpiredNode(now);
-    return new_search_node;
+    return not found;
 }
 
 std::vector<Sp<Node>>
@@ -737,7 +835,7 @@ Dht::Search::getNodes() const
     std::vector<Sp<Node>> ret {};
     ret.reserve(nodes.size());
     for (const auto& sn : nodes)
-        ret.emplace_back(sn.node);
+        ret.emplace_back(sn->node);
     return ret;
 }
 
@@ -746,9 +844,9 @@ Dht::Search::isSynced(time_point now) const
 {
     unsigned i = 0;
     for (const auto& n : nodes) {
-        if (n.isBad())
+        if (n->isBad())
             continue;
-        if (not n.isSynced(now))
+        if (not n->isSynced(now))
             return false;
         if (++i == TARGET_NODES)
             break;
@@ -779,9 +877,9 @@ Dht::Search::isDone(const Get& get) const
 {
     unsigned i = 0;
     for (const auto& sn : nodes) {
-        if (sn.isBad())
+        if (sn->isBad())
             continue;
-        if (not sn.isDone(get))
+        if (not sn->isDone(get))
             return false;
         if (++i == TARGET_NODES)
             break;
@@ -796,9 +894,9 @@ Dht::Search::isAnnounced(Value::Id id) const
         return false;
     unsigned i = 0;
     for (const auto& n : nodes) {
-        if (n.isBad())
+        if (n->isBad())
             continue;
-        if (not n.isAnnounced(id))
+        if (not n->isAnnounced(id))
             return false;
         if (++i == TARGET_NODES)
             return true;
@@ -807,20 +905,20 @@ Dht::Search::isAnnounced(Value::Id id) const
 }
 
 bool
-Dht::Search::isListening(time_point now) const
+Dht::Search::isListening(time_point now, duration listen_expire) const
 {
     if (nodes.empty() or listeners.empty())
         return false;
     unsigned i = 0;
     for (const auto& n : nodes) {
-        if (n.isBad())
+        if (n->isBad())
             continue;
         SearchNode::NodeListenerStatus::const_iterator ls {};
-        for (ls = n.listenStatus.begin(); ls != n.listenStatus.end() ; ++ls) {
-            if (n.isListening(now, ls))
+        for (ls = n->listenStatus.begin(); ls != n->listenStatus.end() ; ++ls) {
+            if (n->isListening(now, ls, listen_expire))
                 break;
         }
-        if (ls == n.listenStatus.end())
+        if (ls == n->listenStatus.end())
             return false;
         if (++i == LISTEN_NODES)
             break;
