@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2014-2017 Savoir-faire Linux Inc.
+ *  Copyright (C) 2014-2020 Savoir-faire Linux Inc.
  *  Authors: Adrien Béraud <adrien.beraud@savoirfairelinux.com>
  *           Simon Désaulniers <simon.desaulniers@savoirfairelinux.com>
  *           Sébastien Blin <sebastien.blin@savoirfairelinux.com>
@@ -17,6 +17,9 @@
  *  You should have received a copy of the GNU General Public License
  *  along with this program. If not, see <https://www.gnu.org/licenses/>.
  */
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 #include "securedht.h"
 #include "rng.h"
@@ -34,7 +37,7 @@ extern "C" {
 namespace dht {
 
 SecureDht::SecureDht(std::unique_ptr<DhtInterface> dht, SecureDht::Config conf)
-: dht_(std::move(dht)), key_(conf.id.first), certificate_(conf.id.second)
+: dht_(std::move(dht)), key_(conf.id.first), certificate_(conf.id.second), enableCache_(conf.cert_cache_all)
 {
     if (!dht_) return;
     for (const auto& type : DEFAULT_TYPES)
@@ -46,7 +49,7 @@ SecureDht::SecureDht(std::unique_ptr<DhtInterface> dht, SecureDht::Config conf)
     registerInsecureType(CERTIFICATE_TYPE);
 
     if (certificate_) {
-        auto certId = certificate_->getPublicKey().getId();
+        auto certId = certificate_->getId();
         if (key_ and certId != key_->getPublicKey().getId())
             throw DhtException("SecureDht: provided certificate doesn't match private key.");
 
@@ -54,15 +57,15 @@ SecureDht::SecureDht(std::unique_ptr<DhtInterface> dht, SecureDht::Config conf)
             CERTIFICATE_TYPE,
             *certificate_,
             1
-        }, [this](bool ok) {
+        }, [this, certId](bool ok) {
             if (ok)
-                DHT_LOG.DBG("SecureDht: public key announced successfully");
+                if (logger_)
+                    logger_->d(certId, "SecureDht: public key announced successfully");
         }, {}, true);
     }
 }
 
-SecureDht::~SecureDht()
-{}
+SecureDht::~SecureDht() = default;
 
 ValueType
 SecureDht::secureType(ValueType&& type)
@@ -74,7 +77,8 @@ SecureDht::secureType(ValueType&& type)
                 v->signatureValid = v->owner and v->owner->checkSignature(v->getToSign(), v->signature);
             }
             if (!v->signatureValid) {
-                DHT_LOG.WARN("Signature verification failed");
+                if (logger_)
+                    logger_->w("Signature verification failed");
                 return false;
             }
         }
@@ -84,7 +88,8 @@ SecureDht::secureType(ValueType&& type)
         if (!o->isSigned())
             return type.editPolicy(id, o, n, nid, a);
         if (o->owner != n->owner) {
-            DHT_LOG.WARN("Edition forbidden: owner changed.");
+            if (logger_)
+                logger_->w("Edition forbidden: owner changed.");
             return false;
         }
         if (!n->signatureChecked) {
@@ -92,14 +97,16 @@ SecureDht::secureType(ValueType&& type)
             n->signatureValid = o->owner and o->owner->checkSignature(n->getToSign(), n->signature);
         }
         if (!n->signatureValid) {
-            DHT_LOG.WARN("Edition forbidden: signature verification failed.");
+            if (logger_)
+                logger_->w("Edition forbidden: signature verification failed.");
             return false;
         }
         if (o->seq == n->seq) {
             // If the data is exactly the same,
             // it can be reannounced, possibly by someone else.
             if (o->getToSign() != n->getToSign()) {
-                DHT_LOG.WARN("Edition forbidden: sequence number must be increasing.");
+                if (logger_)
+                    logger_->w("Edition forbidden: sequence number must be increasing.");
                 return false;
             }
         }
@@ -145,7 +152,8 @@ SecureDht::registerCertificate(const InfoHash& node, const Blob& data)
     }
     InfoHash h = crt->getPublicKey().getId();
     if (node == h) {
-        DHT_LOG.DBG("Registering certificate for %s", h.toString().c_str());
+        if (logger_)
+            logger_->d("Registering certificate for %s", h.toString().c_str());
         auto it = nodesCertificates_.find(h);
         if (it == nodesCertificates_.end())
             std::tie(it, std::ignore) = nodesCertificates_.emplace(h, std::move(crt));
@@ -153,7 +161,8 @@ SecureDht::registerCertificate(const InfoHash& node, const Blob& data)
             it->second = std::move(crt);
         return it->second;
     } else {
-        DHT_LOG.WARN("Certificate %s for node %s does not match node id !", h.toString().c_str(), node.toString().c_str());
+        if (logger_)
+            logger_->w("Certificate %s for node %s does not match node id !", h.toString().c_str(), node.toString().c_str());
         return nullptr;
     }
 }
@@ -166,11 +175,12 @@ SecureDht::registerCertificate(Sp<crypto::Certificate>& cert)
 }
 
 void
-SecureDht::findCertificate(const InfoHash& node, std::function<void(const Sp<crypto::Certificate>)> cb)
+SecureDht::findCertificate(const InfoHash& node, const std::function<void(const Sp<crypto::Certificate>)>& cb)
 {
     Sp<crypto::Certificate> b = getCertificate(node);
     if (b && *b) {
-        DHT_LOG.DBG("Using certificate from cache for %s", node.toString().c_str());
+        if (logger_)
+            logger_->d("Using certificate from cache for %s", node.to_c_str());
         if (cb)
             cb(b);
         return;
@@ -178,7 +188,8 @@ SecureDht::findCertificate(const InfoHash& node, std::function<void(const Sp<cry
     if (localQueryMethod_) {
         auto res = localQueryMethod_(node);
         if (not res.empty()) {
-            DHT_LOG.DBG("Registering certificate from local store for %s", node.toString().c_str());
+            if (logger_)
+                logger_->d("Registering certificate from local store for %s", node.to_c_str());
             nodesCertificates_.emplace(node, res.front());
             if (cb)
                 cb(res.front());
@@ -193,7 +204,8 @@ SecureDht::findCertificate(const InfoHash& node, std::function<void(const Sp<cry
         for (const auto& v : vals) {
             if (auto cert = registerCertificate(node, v->data)) {
                 *found = true;
-                DHT_LOG.DBG("Found certificate for %s", node.toString().c_str());
+                if (logger_)
+                    logger_->d(node, "Found certificate for %s", node.to_c_str());
                 if (cb)
                     cb(cert);
                 return false;
@@ -207,11 +219,12 @@ SecureDht::findCertificate(const InfoHash& node, std::function<void(const Sp<cry
 }
 
 void
-SecureDht::findPublicKey(const InfoHash& node, std::function<void(const Sp<const crypto::PublicKey>)> cb)
+SecureDht::findPublicKey(const InfoHash& node, const std::function<void(const Sp<const crypto::PublicKey>)>& cb)
 {
     auto pk = getPublicKey(node);
     if (pk && *pk) {
-        DHT_LOG.DBG("Found public key from cache for %s", node.toString().c_str());
+        if (logger_)
+            logger_->d(node, "Found public key from cache for %s", node.to_c_str());
         if (cb)
             cb(pk);
         return;
@@ -235,7 +248,7 @@ SecureDht::checkValue(const Sp<Value>& v)
     // Decrypt encrypted values
     if (v->isEncrypted()) {
         if (not key_) {
-#if OPENDHT_PROXY_SERVER
+#ifdef OPENDHT_PROXY_SERVER
             if (forward_all_) // We are currently a proxy, send messages to clients.
                 return v;
 #endif
@@ -255,7 +268,8 @@ SecureDht::checkValue(const Sp<Value>& v)
             }
             // Ignore values belonging to other people
         } catch (const std::exception& e) {
-            DHT_LOG.WARN("Could not decrypt value %s : %s", v->toString().c_str(), e.what());
+            if (logger_)
+                logger_->w("Could not decrypt value %s : %s", v->toString().c_str(), e.what());
         }
     }
     // Check signed values
@@ -266,11 +280,12 @@ SecureDht::checkValue(const Sp<Value>& v)
         v->signatureChecked = true;
         if (v->owner and v->owner->checkSignature(v->getToSign(), v->signature)) {
             v->signatureValid = true;
-            nodesPubKeys_[v->owner->getId()] = v->owner;
+            if (enableCache_)
+                nodesPubKeys_[v->owner->getId()] = v->owner;
             return v;
         }
-        else
-            DHT_LOG.WARN("Signature verification failed for %s", v->toString().c_str());
+        else if (logger_)
+            logger_->w("Signature verification failed for %s", v->toString().c_str());
     }
     // Forward normal values
     else {
@@ -280,10 +295,12 @@ SecureDht::checkValue(const Sp<Value>& v)
 }
 
 ValueCallback
-SecureDht::getCallbackFilter(ValueCallback cb, Value::Filter&& filter)
+SecureDht::getCallbackFilter(const ValueCallback& cb, Value::Filter&& filter)
 {
     return [=](const std::vector<Sp<Value>>& values, bool expired) {
         std::vector<Sp<Value>> tmpvals {};
+        if (not filter)
+            tmpvals.reserve(values.size());
         for (const auto& v : values) {
             if (auto nv = checkValue(v))
                 if (not filter or filter(*nv))
@@ -297,10 +314,12 @@ SecureDht::getCallbackFilter(ValueCallback cb, Value::Filter&& filter)
 
 
 GetCallback
-SecureDht::getCallbackFilter(GetCallback cb, Value::Filter&& filter)
+SecureDht::getCallbackFilter(const GetCallback& cb, Value::Filter&& filter)
 {
     return [=](const std::vector<Sp<Value>>& values) {
         std::vector<Sp<Value>> tmpvals {};
+        if (not filter)
+            tmpvals.reserve(values.size());
         for (const auto& v : values) {
             if (auto nv = checkValue(v))
                 if (not filter or filter(*nv))
@@ -334,6 +353,11 @@ SecureDht::listen(const InfoHash& id, GetCallback cb, Value::Filter f, Where w)
 void
 SecureDht::putSigned(const InfoHash& hash, Sp<Value> val, DoneCallback callback, bool permanent)
 {
+    if (not key_)  {
+        if (callback)
+            callback(false, {});
+        return;
+    }
     if (val->id == Value::INVALID_ID) {
         crypto::random_device rdev;
         std::uniform_int_distribution<Value::Id> rand_id;
@@ -349,13 +373,16 @@ SecureDht::putSigned(const InfoHash& hash, Sp<Value> val, DoneCallback callback,
     // Check if data already exists on the dht
     get(hash,
         [val,this] (const std::vector<Sp<Value>>& vals) {
-            DHT_LOG.DBG("Found online previous value being announced.");
+            if (logger_)
+                logger_->d("Found online previous value being announced.");
             for (const auto& v : vals) {
-                if (!v->isSigned())
-                    DHT_LOG.ERR("Existing non-signed value seems to exists at this location.");
-                else if (not v->owner or v->owner->getId() != getId())
-                    DHT_LOG.ERR("Existing signed value belonging to someone else seems to exists at this location.");
-                else if (val->seq <= v->seq)
+                if (!v->isSigned()) {
+                    if (logger_)
+                        logger_->e("Existing non-signed value seems to exists at this location.");
+                } else if (not v->owner or v->owner->getId() != getId()) {
+                    if (logger_)
+                        logger_->e("Existing signed value belonging to someone else seems to exists at this location.");
+                } else if (val->seq <= v->seq)
                     val->seq = v->seq + 1;
             }
             return true;
@@ -372,17 +399,24 @@ SecureDht::putSigned(const InfoHash& hash, Sp<Value> val, DoneCallback callback,
 void
 SecureDht::putEncrypted(const InfoHash& hash, const InfoHash& to, Sp<Value> val, DoneCallback callback, bool permanent)
 {
+    if (not key_)  {
+        if (callback)
+            callback(false, {});
+        return;
+    }
     findPublicKey(to, [=](const Sp<const crypto::PublicKey>& pk) {
         if(!pk || !*pk) {
             if (callback)
                 callback(false, {});
             return;
         }
-        DHT_LOG.WARN("Encrypting data for PK: %s", pk->getId().toString().c_str());
+        if (logger_)
+            logger_->w("Encrypting data for PK: %s", pk->getId().toString().c_str());
         try {
             dht_->put(hash, encrypt(*val, *pk), callback, time_point::max(), permanent);
         } catch (const std::exception& e) {
-            DHT_LOG.ERR("Error putting encrypted data: %s", e.what());
+            if (logger_)
+                logger_->e("Error putting encrypted data: %s", e.what());
             if (callback)
                 callback(false, {});
         }
